package com.emamaker.voxeltest.intervaltrees.world;

import java.util.Random;
import java.util.concurrent.ConcurrentHashMap;

import com.emamaker.voxeltest.intervaltrees.Voxel;
import com.emamaker.voxeltest.intervaltrees.utils.OpenSimplexNoise;
import com.emamaker.voxeltest.intervaltrees.world.chunk.ChunkGenerator;
import com.emamaker.voxeltest.intervaltrees.world.chunk.ChunkMesher;
import com.emamaker.voxeltest.intervaltrees.utils.Config;
import com.emamaker.voxeltest.intervaltrees.utils.SpaceFillingCurve;
import com.emamaker.voxeltest.intervaltrees.utils.Utils;
import com.emamaker.voxeltest.intervaltrees.world.chunk.Chunk;
import com.jme3.scene.Node;

public class WorldManager {

    public final Voxel game;

    public int camx, camy, camz;

    ChunkMesher chunkMesher = new ChunkMesher();
    ChunkGenerator chunkGenerator = new ChunkGenerator();
    OpenSimplexNoise noiseGenerator;

    Node worldNode = new Node();

    ConcurrentHashMap<Integer, Chunk> chunks = new ConcurrentHashMap<>();

    public WorldManager(final Voxel game_) {
        this.game = game_;
    }

    public void initWorld() {
        noiseGenerator = new OpenSimplexNoise(new Random().nextLong());
    }

    int chunkX, chunkY, chunkZ;
    int index;
    public boolean generated = false, meshed = false, worldGenerated = false;

    public void update() {
        camx = (int) (game.pos.getX());
        camy = (int) (game.pos.getY());
        camz = (int) (game.pos.getZ());

        // clamp to "chunk" coordinates
        // The chunk with origin at x,y,z in chunk coords actually has the origin at
        // x*chunksize, y*chunksize, z*chunksize in world coords
        chunkX = (int) (camx / Config.CHUNK_LENGTH);
        chunkY = (int) (camy / Config.CHUNK_LENGTH);
        chunkZ = (int) (camz / Config.CHUNK_LENGTH);

        for (int i = Math.max(0, chunkX - Config.RENDER_DISTANCE); i < chunkX + Config.RENDER_DISTANCE; i++) {
            for (int j = Math.max(0, chunkY - Config.RENDER_DISTANCE); j < chunkY + Config.RENDER_DISTANCE; j++) {
                for (int k = Math.max(0, chunkZ - Config.RENDER_DISTANCE); k < chunkZ + Config.RENDER_DISTANCE; k++) {
                    if(!Utils.withinDistance(chunkX, chunkY, chunkZ, i,j,k,Config.RENDER_DISTANCE)) continue;

                    index = SpaceFillingCurve.Morton_3D_Encode_10bit((short)i, (short)j, (short)k);
                    if (!chunks.containsKey(index)) {
                        chunks.put(index, new Chunk(i,j,k));
                        worldGenerated = false;
                    }
                }
            }
        }

        generated = false;
        meshed = false;
        for(Chunk c : chunks.values()){
           if(!generated && !c.bgetState(Chunk.CHUNK_STATE_GENERATED)) {
               chunkGenerator.generateNoise(c);
               c.setState(Chunk.CHUNK_STATE_GENERATED, true);
               generated = true;
           }
            if(!meshed && c.bgetState(Chunk.CHUNK_STATE_GENERATED)&& !c.bgetState(Chunk.CHUNK_STATE_MESHED)) {
                chunkMesher.greedyMeshing(this, c);
                c.setState(Chunk.CHUNK_STATE_MESHED, true);
                meshed = true;
            }
        }

        if(!generated && !meshed && !worldGenerated){
            worldGenerated = true;
            System.out.println("Finished world generation");
        }

    }
    public void render(){
        for(Chunk c : chunks.values()){
            if(!c.bgetState(Chunk.CHUNK_STATE_EMPTY) && !c.bgetState(Chunk.CHUNK_STATE_LOADED) && c.bgetState(Chunk.CHUNK_STATE_GENERATED) && c.bgetState(Chunk.CHUNK_STATE_MESHED)){
                game.getRootNode().attachChild(c.chunkNode);
                c.setState(Chunk.CHUNK_STATE_LOADED, true);
            }
            if(!Utils.withinDistance(chunkX, chunkY, chunkZ, c.posx, c.posy, c.posz, Config.RENDER_DISTANCE)){
                c.chunkNode.removeFromParent();
                chunks.remove(SpaceFillingCurve.Morton_3D_Encode_10bit((short)c.posx, (short)c.posy, (short)c.posz));
            }
        }
    }

    public boolean chunkInCameraLimits(Chunk c) {
        return c.posx >= Math.min(0, chunkX - Config.RENDER_DISTANCE ) && c.posx < Math.min(0, chunkX + Config.RENDER_DISTANCE) && c.posy >= chunkY - Config.RENDER_DISTANCE && c.posy < Math.min(0, chunkY + Config.RENDER_DISTANCE) && c.posz >= chunkZ - Config.RENDER_DISTANCE && c.posz < Math.min(0, chunkZ + Config.RENDER_DISTANCE);
    }

    public void benchmark(){
        for(int i = 0; i < Config.RENDER_DISTANCE*2; i++)
            for(int j = 0; j < Config.RENDER_DISTANCE*2; j++)
                for(int k = 0; k < Config.RENDER_DISTANCE*2; k++){
                    Chunk c = new Chunk(i,j,k);
                    chunkGenerator.arrayGenerateCorner(c);
                    chunkMesher.greedyMeshing(this, c);
                    game.getRootNode().attachChild(c.chunkNode);
                }
    }

}
